@@ -2,20 +2,22 @@
 shapefile.py
 Provides read and write support for ESRI Shapefiles.
 author: jlawhead<at>geospatialpython.com
-date: 20140507
-version: 1.2.1
+date: 2016/09/24
+version: 1.2.10
 Compatible with Python versions 2.4-3.x
-version changelog: Fixed u() to just return the byte sequence on exception 
+version changelog: 
+- Bump version to fix pip install issue.
 """
 
-__version__ = "1.2.1"
+__version__ = "1.2.10"
 
-from struct import pack, unpack, calcsize, error
+from struct import pack, unpack, calcsize, error, Struct
 import os
 import sys
 import time
 import array
 import tempfile
+import itertools
 
 #
 # Constants for shape types
@@ -38,6 +40,9 @@ PYTHON3 = sys.version_info[0] == 3
 
 if PYTHON3:
     xrange = range
+    izip = zip
+else:
+    from itertools import izip
 
 def b(v):
     if PYTHON3:
@@ -92,8 +97,7 @@ class _Array(array.array):
 
 def signed_area(coords):
     """Return the signed area enclosed by a ring using the linear time
-    algorithm at http://www.cgafaq.info/wiki/Polygon_Area. A value >= 0
-    indicates a counter-clockwise oriented ring.
+    algorithm. A value >= 0 indicates a counter-clockwise oriented ring.
     """
     xs, ys = map(list, zip(*coords))
     xs.append(xs[1])
@@ -333,7 +337,7 @@ class Reader:
         if shapeType in (3,5,13,15,23,25,31):
             nParts = unpack("<i", f.read(4))[0]
         # Shape types with points
-        if shapeType in (3,5,8,13,15,23,25,31):
+        if shapeType in (3,5,8,13,15,18,23,25,28,31):
             nPoints = unpack("<i", f.read(4))[0]
         # Read parts
         if nParts:
@@ -343,7 +347,8 @@ class Reader:
             record.partTypes = _Array('i', unpack("<%si" % nParts, f.read(nParts * 4)))
         # Read points - produces a list of [x,y] values
         if nPoints:
-            record.points = [_Array('d', unpack("<2d", f.read(16))) for p in range(nPoints)]
+            flat = unpack("<%sd" % (2 * nPoints), f.read(16*nPoints))
+            record.points = list(izip(*(iter(flat),) * 2))
         # Read z extremes and values
         if shapeType in (13,15,18,31):
             (zmin, zmax) = unpack("<2d", f.read(16))
@@ -386,10 +391,12 @@ class Reader:
             numRecords = shxRecordLength // 8
             # Jump to the first record.
             shx.seek(100)
-            for r in range(numRecords):
-                # Offsets are 16-bit words just like the file length
-                self._offsets.append(unpack(">i", shx.read(4))[0] * 2)
-                shx.seek(shx.tell() + 4)
+            shxRecords = _Array('i')
+            # Each offset consists of two nrs, only the first one matters
+            shxRecords.fromfile(shx, 2 * numRecords)
+            if sys.byteorder != 'big':
+                 shxRecords.byteswap()
+            self._offsets = [2 * el for el in shxRecords[::2]]
         if not i == None:
             return self._offsets[i]
 
@@ -463,11 +470,14 @@ class Reader:
             fieldDesc[1] = u(fieldDesc[1])
             self.fields.append(fieldDesc)
         terminator = dbf.read(1)
-        assert terminator == b("\r")
+        if terminator != b("\r"):
+            raise ShapefileException("Shapefile dbf header lacks expected terminator. (likely corrupt?)")
         self.fields.insert(0, ('DeletionFlag', 'C', 1, 0))
+        fmt,fmtSize = self.__recordFmt()
+        self.__recStruct = Struct(fmt)
 
     def __recordFmt(self):
-        """Calculates the size of a .shp geometry record."""
+        """Calculates the format and size of a .dbf record."""
         if not self.numRecords:
             self.__dbfHeader()
         fmt = ''.join(['%ds' % fieldinfo[2] for fieldinfo in self.fields])
@@ -477,8 +487,7 @@ class Reader:
     def __record(self):
         """Reads and returns a dbf record row as a list of values."""
         f = self.__getFileObj(self.dbf)
-        recFmt = self.__recordFmt()
-        recordContents = unpack(recFmt[0], f.read(recFmt[1]))
+        recordContents = self.__recStruct.unpack(f.read(self.__recStruct.size))
         if recordContents[0] != b(' '):
             # deleted record
             return None
@@ -492,18 +501,30 @@ class Reader:
                 continue
             elif typ == "N":
                 value = value.replace(b('\0'), b('')).strip()
+                value = value.replace(b('*'), b(''))  # QGIS NULL is all '*' chars
                 if value == b(''):
-                    value = 0
+                    value = None
                 elif deci:
-                    value = float(value)
+                    try:
+                        value = float(value)
+                    except ValueError:
+                        #not parseable as float, set to None
+                        value = None
                 else:
-                    value = int(value)
+                    try:
+                        value = int(value)
+                    except ValueError:
+                        #not parseable as int, set to None
+                        value = None
             elif typ == b('D'):
-                try:
-                    y, m, d = int(value[:4]), int(value[4:6]), int(value[6:8])
-                    value = [y, m, d]
-                except:
-                    value = value.strip()
+                if value.count(b('0')) == len(value):  # QGIS NULL is all '0' chars
+                    value = None
+                else:
+                    try:
+                        y, m, d = int(value[:4]), int(value[4:6]), int(value[6:8])
+                        value = [y, m, d]
+                    except:
+                        value = value.strip()
             elif typ == b('L'):
                 value = (value in b('YyTt') and b('T')) or \
                                         (value in b('NnFf') and b('F')) or b('?')
@@ -519,7 +540,7 @@ class Reader:
         if not self.numRecords:
             self.__dbfHeader()
         i = self.__restrictIndex(i)
-        recSize = self.__recordFmt()[1]
+        recSize = self.__recStruct.size
         f.seek(0)
         f.seek(self.__dbfHeaderLength() + (i * recSize))
         return self.__record()
@@ -562,6 +583,13 @@ class Reader:
         return [_ShapeRecord(shape=rec[0], record=rec[1]) \
                                 for rec in zip(self.shapes(), self.records())]
 
+    def iterShapeRecords(self):
+        """Returns a generator of combination geometry/attribute records for
+        all records in a shapefile."""
+        for shape, record in izip(self.iterShapes(), self.iterRecords()):
+            yield _ShapeRecord(shape=shape, record=record)
+
+
 class Writer:
     """Provides write support for ESRI Shapefiles."""
     def __init__(self, shapeType=None):
@@ -898,7 +926,10 @@ class Writer:
                     value = str(value)[0].upper()
                 else:
                     value = str(value)[:size].ljust(size)
-                assert len(value) == size
+                if len(value) != size:
+                    raise ShapefileException(
+                        "Shapefile Writer unable to pack incorrect sized value"
+                        " (size %d) into field '%s' (size %d)." % (len(value), fieldName, size))
                 value = b(value)
                 f.write(value)
 
@@ -1164,11 +1195,11 @@ class Editor(Writer):
 def test():
     import doctest
     doctest.NORMALIZE_WHITESPACE = 1
-    doctest.testfile("README.txt", verbose=1)
+    doctest.testfile("README.md", verbose=1)
 
 if __name__ == "__main__":
     """
-    Doctests are contained in the file 'README.txt'. This library was originally developed
+    Doctests are contained in the file 'README.md'. This library was originally developed
     using Python 2.3. Python 2.4 and above have some excellent improvements in the built-in
     testing libraries but for now unit testing is done using what's available in
     2.3.
